package memcached

import (
	"context"
	"flag"
	"fmt"
	"regexp"
	"strconv"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
)

var log = _log.New("memcached")

var IGNORE_METRICS = []string{
	"pid",
	"time",
	"version",
	"libevent",
	"pointer_size",
	"rusage_user",
	"rusage_system",
	"connection_structures",
	"reserved_fds",
	"hash_power_level",
	"lru_maintainer_juggles",
	"lrutail_reflocked",
	"lru_bumps_dropped",
}

var regexCacheCommand = regexp.MustCompile(`^(get|delete|incr|decr|cas|touch)_(hits|misses|badval|expired|flushed)$`)

var confTimeout = 3 * time.Second
var confInterval = time.Duration(0)

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confTimeout, "memcached.timeout", confTimeout, "[memcached] operation timeout")
	fs.DurationVar(&confInterval, "memcached.interval", confInterval, "[memcached] interval between polls")
}

var Factory = newFactory()

type memcachedDaemonFactory struct {
	*daemon.TcpDaemonFactory
}

func newFactory() *memcachedDaemonFactory {
	tcpDaemonFactory := daemon.NewTcpDaemonFactory("memcached").(*daemon.TcpDaemonFactory)

	return &memcachedDaemonFactory{
		TcpDaemonFactory: tcpDaemonFactory,
	}
}

func (df *memcachedDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	tcpDaemon := df.TcpDaemonFactory.New(paramsI).(*daemon.TcpDaemon)
	tcpDaemon.Timeout = confTimeout
	tcpDaemon.Interval_ = confInterval
	tcpDaemon.Keepalive = true
	tcpDaemon.Logger = log

	return &memcachedDaemon{
		TcpDaemon: tcpDaemon,
	}
}

type memcachedDaemon struct {
	*daemon.TcpDaemon
}

func (d *memcachedDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	if err := d.collectStats(ctx, m); err != nil {
		m.IncrError("memcached", "memcached")
		return fmt.Errorf("Error collecting memcached '%s' metrics: %w", d.ServiceID, err)
	}
	if err := d.collectSlabs(ctx, m); err != nil {
		m.IncrError("memcached", "memcached")
		return fmt.Errorf("Error collecting memcached '%s' metrics: %w", d.ServiceID, err)
	}

	return nil
}

func (d *memcachedDaemon) collectStats(ctx context.Context, m *common.MetricData) error {
	data, err := d.Command(ctx, "stats\r\n", []string{"END", ""})
	if err != nil {
		return err
	}

	for _, line := range data {
		kv := strings.SplitN(line, " ", 3)
		if len(kv) != 3 || kv[0] != "STAT" {
			// TODO log
			continue
		}
		stat := kv[1]
		valStr := kv[2]
		if common.Index(IGNORE_METRICS, stat) > -1 {
			continue
		}

		labels := map[string]string{
			"namespace":    d.Namespace,
			"service":      d.ServiceID,
			"service_type": "memcached",
			"service_addr": d.ServiceAddr,
		}

		val, errFloat := strconv.ParseFloat(valStr, 64)

		/* match commands get_hits, get_misses, delete_hits, ... */
		matches := regexCacheCommand.FindStringSubmatch(stat)
		if len(matches) == 3 {
			stat = "cache_commands_total"
			labels["command"] = matches[1]
			labels["cache"] = matches[2]

		} else {

			switch stat {
			case "max_connections":
				stat = "connections_max"
			case "curr_connections":
				stat = "connections"
			case "total_connections":
				stat = "connections_total"

			case "cmd_get":
				stat = "commands_total"
				labels["command"] = "get"
			case "cmd_set":
				stat = "commands_total"
				labels["command"] = "set"
			case "cmd_flush":
				stat = "commands_total"
				labels["command"] = "flush"
			case "cmd_touch":
				stat = "commands_total"
				labels["command"] = "touch"
			case "auth_cmds":
				stat = "commands_total"
				labels["command"] = "auth"
			case "auth_errors":
				stat = "commands_total"
				labels["command"] = "auth_error"

			case "bytes_read":
				stat = "io_bytes_total"
				labels["direction"] = "in"
			case "bytes_written":
				stat = "io_bytes_total"
				labels["direction"] = "out"

			case "listen_disabled_num":
				stat = "listen_disabled_total"
			case "time_in_listen_disabled_us":
				if errFloat != nil {
					continue
				}
				stat = "listen_disabled_duration_seconds"
				val /= 10000 // convert µs to seconds

			case "conn_yields":
				stat += "_total"

			case "slab_reassign_rescues":
				stat = "slab_reassigns"
				labels["type"] = "rescue"
			case "slab_reassign_chunk_rescues":
				stat = "slab_reassigns"
				labels["type"] = "chunk_rescue"
			case "slab_reassign_evictions_nomem":
				stat = "slab_reassigns"
				labels["type"] = "eviction_nomem"
			case "slab_reassign_inline_reclaim":
				stat = "slab_reassigns"
				labels["type"] = "inline_reclaim"
			case "slab_reassign_busy_items":
				stat = "slab_reassigns"
				labels["type"] = "busy_item"
			case "slab_reassign_busy_deletes":
				stat = "slab_reassigns"
				labels["type"] = "busy_delete"

			case "slabs_moved":
				stat += "_total"

			case "log_worker_dropped":
				stat = "log_workers_total"
				labels["state"] = "dropped"
			case "log_worker_written":
				stat = "log_workers_total"
				labels["state"] = "written"
			case "log_watcher_skipped":
				stat = "log_watchers_total"
				labels["state"] = "skipped"
			case "log_watcher_sent":
				stat = "log_watchers_total"
				labels["state"] = "sent"

			case "bytes":
				stat = "size_bytes"
			case "limit_maxbytes":
				stat = "size_bytes_max"

			case "hash_bytes":
				stat = "hash_size_bytes"

			case "curr_items":
				stat = "stored_items"
			case "total_items":
				stat = "stored_items_total"

			case "lru_crawler_starts":
				stat += "_total"
			case "crawler_reclaimed":
				stat = "lru_" + stat + "_total"
			case "crawler_items_checked":
				stat = "lru_" + stat + "_total"

			case "moves_to_cold":
				stat = "lru_moves_total"
				labels["type"] = "to_cold"
			case "moves_to_warm":
				stat = "lru_moves_total"
				labels["type"] = "to_warm"
			case "moves_within_lru":
				stat = "lru_moves_total"
				labels["type"] = "within_lru"

			case "direct_reclaims":
				stat += "_total"

			case "uptime":
				stat += "_seconds"

				//			case "expired_unfetched":

			case "evictions":
				stat = "lru_evictions_total"
				labels["type"] = "valid"
			case "evicted_unfetched":
				stat = "lru_evictions_total"
				labels["type"] = "unfetched"
			case "evicted_active":
				stat = "lru_evictions_total"
				labels["type"] = "active"

			case "expired_unfetched":
				stat = "lru_expires_total"
				labels["type"] = "unfetched"

			case "reclaimed":
				stat = "lru_reclaimed_total"

			/* by default use the metrics as-is and consider it's a float */
			default:
				/* non exhaustive list:
				 *   accepting_conns
				 *   threads
				 *   hash_is_expanding
				 *   slab_reassign_running
				 *   lru_crawler_running
				 *   malloc_fails
				 *   slab_global_page_pool
				 *   rejected_connections
				 */
			}
		}
		if errFloat != nil {
			// TODO log
			continue
		}
		m.Set(&common.Metric{
			Name:   "memcached_" + stat,
			Labels: labels,
			Value:  val,
			TTL:    confInterval * 2,
		})
	}

	return nil
}

func (d *memcachedDaemon) collectSlabs(ctx context.Context, m *common.MetricData) error {
	data, err := d.Command(ctx, "stats slabs\r\n", []string{"END", ""})
	if err != nil {
		return err
	}

	total_chunks := make(map[string]float64)
	chunk_size := make(map[string]float64)
	mem_requested := make(map[string]float64)

	for _, line := range data {
		kv := strings.SplitN(line, " ", 3)
		if len(kv) != 3 || kv[0] != "STAT" {
			// TODO log
			continue
		}
		stat := kv[1]
		valStr := kv[2]

		val, errFloat := strconv.ParseFloat(valStr, 64)
		if errFloat != nil {
			// TODO log
			continue
		}

		labels := map[string]string{
			"namespace":    d.Namespace,
			"service":      d.ServiceID,
			"service_type": "memcached",
			"service_addr": d.ServiceAddr,
		}

		if stat == "active_slabs" {
			m.Set(&common.Metric{
				Name:   "memcached_slab_actives_total",
				Labels: labels,
				Value:  val,
				TTL:    confInterval * 2,
			})
			continue
		}

		if stat == "total_malloced" {
			m.Set(&common.Metric{
				Name:   "memcached_slab_memory_size_bytes_total",
				Labels: labels,
				Value:  val,
				TTL:    confInterval * 2,
			})
			continue
		}

		matches := strings.SplitN(stat, ":", 2)
		if len(matches) != 2 {
			continue
		}
		slab := matches[0]
		stat = matches[1]
		labels["slab"] = slab

		switch stat {
		case "total_pages":
			stat = "pages_total"
		case "total_chunks":
			stat = "chunks_total"
			total_chunks[slab] = val
		case "chunk_size":
			stat += "_bytes"
			chunk_size[slab] = val
		case "mem_requested":
			stat += "_bytes"
			mem_requested[slab] = val

		case "get_hits":
			stat = "commands_total"
			labels["command"] = "get"
		case "cmd_set":
			stat = "commands_total"
			labels["command"] = "set"
		case "delete_hits":
			stat = "commands_total"
			labels["command"] = "delete"
		case "incr_hits":
			stat = "commands_total"
			labels["command"] = "incr"
		case "decr_hits":
			stat = "commands_total"
			labels["command"] = "decr"
		case "cas_hits":
			stat = "commands_total"
			labels["command"] = "cas"
		case "cas_badval":
			stat = "commands_total"
			labels["command"] = "cas_badval"
		case "touch_hits":
			stat = "commands_total"
			labels["command"] = "touch"

		case "used_chunks":
			stat = "chunks_total"
			labels["type"] = "used"
		case "free_chunks":
			stat = "chunks_total"
			labels["type"] = "free"

		default:
			/*
			 *  chunks_per_page
			 *  free_chunks_end
			 */
		}

		m.Set(&common.Metric{
			Name:   "memcached_slab_" + stat,
			Labels: labels,
			Value:  val,
			TTL:    confInterval * 2,
		})
	}

	/*
	 * from https://github.com/memcached/memcached/blob/master/doc/protocol.txt
	 * Items are stored in a slab that is the same size or larger than the
	 * item.  mem_requested shows the size of all items within a
	 * slab. (total_chunks * chunk_size) - mem_requested shows memory
	 * wasted in a slab class.  If you see a lot of waste, consider tuning
	 * the slab factor.
	 */
	for slab, total := range total_chunks {
		chunk, ok := chunk_size[slab]
		if !ok {
			continue
		}
		requested, ok := mem_requested[slab]
		if !ok {
			continue
		}
		m.Set(&common.Metric{
			Name: "memcached_slab_wasted_bytes",
			Labels: map[string]string{
				"namespace":    d.Namespace,
				"service":      d.ServiceID,
				"service_type": "memcached",
				"service_addr": d.ServiceAddr,
				"slab":         slab,
			},
			Value: (total * chunk) - requested,
			TTL:   confInterval * 2,
		})
	}

	return nil
}
